#include "stdafx.h"

#include "AlbumCanvas.h"
#include "AlbumView.h"
#include "AlbumPrefetch.h"
#include "AlbumController.h"

#include "Browser.h"
#include "Persistence.h"

#include "../images/bitmaps.h"

#include <sstream>
#include <iostream>
#include <iomanip>
#include <cmath>

#include <wx/window.h>
#include <wx/mstream.h>
#include <wx/dcclient.h>
#include <wx/image.h>
#include <wx/dirdlg.h> 
#include <wx/tooltip.h>

// When using GTK, we actually completely bypass wx, gtk and gdk
// to do the scrolling: we directly do the XCopyArea or the guffaw 
// ourselves. This is due to the lack of a per-region scrolling API
// to only scroll the albums, and not the soft scrollbar
#ifdef USE_GTK_PIXBUF
#include <wx/gtk/win_gtk.h>
#include <gdk/gdk.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#endif

#include "AlbumFader.h"


static const int ID_TIMER_SCROLL = 1111;
const int AlbumCanvas::thumbnailSizes[] = {48, 80, 112, 160, 224 }; 


BEGIN_EVENT_TABLE(AlbumCanvas, wxWindow)
  EVT_PAINT(AlbumCanvas::OnPaint)
  EVT_ERASE_BACKGROUND(AlbumCanvas::OnEraseBackGround)
  EVT_TIMER(ID_TIMER_SCROLL, AlbumCanvas::OnScrollTimer)
  EVT_LEFT_DOWN(AlbumCanvas::OnClick)
  EVT_LEFT_UP(AlbumCanvas::OnClick)
  EVT_LEFT_DCLICK(AlbumCanvas::OnDoubleClick)
  EVT_RIGHT_DOWN(AlbumCanvas::OnRightClick)
  EVT_SIZE(AlbumCanvas::OnResize)
  EVT_MOTION(AlbumCanvas::MouseMotion)
END_EVENT_TABLE()

AlbumCanvas::Activity m_activity;

static int s_thumb = 0;
static int s_scroll_start_y;
static int s_timer = 0;
static int s_distance = 0;
static int currentIndex =0 ;

// auto-scroll
static double s_step = 7;
static double s_pos = 0;


AlbumCanvas::AlbumCanvas( wxWindow *parent, GaugeControler * gg, 
						 wxWindowID id, const wxPoint &pos, const wxSize &size )
        : wxWindow( parent, id, pos, size )
{
	m_pCurrentFile = 0;
	m_localScrollPosition = 0;
	m_currentHeight = 0;
	m_pAlbumView = 0;

    SetBackgroundColour(* wxWHITE);
    m_pGaugeControler = gg; 
    m_pTimer = new wxTimer(this, ID_TIMER_SCROLL);
	SetDropTarget(new AlbumDropTarget(this));
	browser::Browser::Instance();// kick-start the browsing prefetch thread

}

AlbumCanvas::~AlbumCanvas()
{
	delete m_pTimer;
	AlbumPrefetch::Instance()->Stop();
	browser::Browser::Instance()->Stop();
}

void AlbumCanvas::SetController(AlbumController * pController) 
{ 
    m_pController=pController; 
	PushEventHandler(m_pController);
}


//////////////////////////////////////////////////////////////////////////
// Album-related events coming from the Main Frame
//////////////////////////////////////////////////////////////////////////

void AlbumCanvas::SetAlbums(const std::vector <browser::AlbumInfo * > & albums)
{
	for (size_t i=0; i <m_albums.size(); ++i)
	{
        m_albums[i]->DiscardAlbum();
		delete m_albums[i];
	}
	m_albums.clear();

	// Just a hack to force the start of the prefetch-thread
	AlbumPrefetch::Instance();

	m_currentHeight = m_localScrollPosition = 0;

	for (size_t i=0; i<albums.size(); ++i)
	{
		if (!albums[i]->entries.empty())
		{
			AlbumView * pAlbumView = new AlbumView(this, albums[i]);
			pAlbumView->screen_pos = m_currentHeight;
			m_currentHeight += pAlbumView->GetHeight();
			m_albums.push_back(pAlbumView);
		}
	}
	Refresh();
}

void AlbumCanvas::SelectAlbum(browser::AlbumInfo * pAlbum)
{
	for (size_t i=0; i<m_albums.size(); ++i)
	{
		if (m_albums[i]->GetAlbumInfo() == pAlbum)
		{
			JumpTo(m_albums[i]->screen_pos);
			return;
		}
	}

		
}

void AlbumCanvas::SetAlbumsThumbnailSize(int x)
{
	double ratio = ((double)m_localScrollPosition)/(m_currentHeight);
	m_currentHeight =  0;

	for (size_t i=0; i<m_albums.size(); ++i)
	{
		AlbumView * pAlbum = m_albums[i];
		pAlbum->screen_pos = m_currentHeight;
		pAlbum->DiscardAlbum();
		pAlbum->SetThumbnailSize(x);
		m_currentHeight += pAlbum->GetHeight();
	}
	Refresh();
	m_localScrollPosition = int(ratio * m_currentHeight);
}


//////////////////////////////////////////////////////////////////////////
// Selection Handling
//////////////////////////////////////////////////////////////////////////
wxDragResult AlbumDropTarget::OnDragOver(wxCoord x, wxCoord y, wxDragResult def)
{
	AlbumView * pAlbum;
	browser::AlbumEntry * pEntry;
	wxMouseEvent e;
	e.m_x=x;
	e.m_y=y;
	m_pCanvas->GetEntry(e, pAlbum, pEntry);

	if (pEntry && def!=wxDragNone)
		return wxDragCopy;
	else
		return wxDragNone;
}

bool AlbumDropTarget::OnDropText(wxCoord x, wxCoord y, const wxString& data)
{
	AlbumView * pAlbum;
	browser::AlbumEntry * pEntry;
	wxMouseEvent e;
	e.m_x=x;
	e.m_y=y;
	m_pCanvas->GetEntry(e, pAlbum, pEntry);
	
	if (pEntry)
	{
		pEntry->pFileEntry->tagged = true; // common::Debug(data.ToAscii());
		std::cout << "Added the tag " << data << " to the file " << pEntry->pFileEntry->name << std::endl;
		m_pCanvas->Refresh(); // TODO: once we have the API to do it, just refresh the impacted picture(s)
	}
	return pEntry!=0;
}

const browser::ClipboardType & AlbumCanvas::GetSelection() const
{
	return m_clipboard;
}

void AlbumCanvas::SelectionClick(wxMouseEvent& event)
{
	AlbumView * pAlbum;
	browser::AlbumEntry * pEntry;
	GetEntry(event, pAlbum, pEntry);

	if (!pEntry)
		return;

	if (!event.m_controlDown && !event.m_shiftDown)
	{
		browser::ClipboardType::iterator it;
		for (it=m_clipboard.begin(); it!=m_clipboard.end(); ++it)
			(*it)->selected=false;
		m_clipboard.clear();
		m_clipboard.insert(pEntry);
		m_pCurrentFile=pEntry;
		m_pAlbumView = pAlbum;
		pEntry->selected=true;
	}
	else if (event.m_controlDown)
	{
		if (pEntry->selected)
		{
			browser::ClipboardType::iterator it = m_clipboard.find(pEntry);
			m_clipboard.erase(it);
			m_pCurrentFile= 0;
		}
		else
		{
			m_clipboard.insert(pEntry);
			m_pCurrentFile=pEntry;
			m_pAlbumView = pAlbum;
		}
		pEntry->selected=!pEntry->selected;
	}
	else if (event.m_shiftDown)
	{
	}
	
	m_pController->UpdateSelection();

	Refresh();
}

browser::FileEntry * AlbumCanvas::GetCurrentFile()
{ 
	if (m_pCurrentFile)
		return m_pCurrentFile->pFileEntry;
	else if (m_clipboard.empty())
		return 0;
	else
		return ((*m_clipboard.begin())->pFileEntry);
}


void AlbumCanvas::NextPicture(bool previous)
{
	if (m_clipboard.size()>1)
	{

		browser::ClipboardType::const_iterator 
			it = m_clipboard.begin(),
			itEnd=m_clipboard.end();

		int idx=0;
		while (it!=itEnd && idx!=currentIndex)
			it++, idx++;

		int id = 1;
		if (previous)
			id = m_clipboard.size()-1;

		currentIndex=(currentIndex+id)%m_clipboard.size();
		while (id--)
		{
			if (++it==itEnd)
				it=m_clipboard.begin();
		}

		m_pCurrentFile = *it;
	}
	else if (m_pAlbumView)
	{
		const AlbumView::Entries  & entries = m_pAlbumView->m_entries;

		if (previous)
			currentIndex = currentIndex+entries.size()-1;
		else
			currentIndex++;

		m_pCurrentFile = entries[(currentIndex)%entries.size()];
	}
	else
		BREAKPOINT;

}


//////////////////////////////////////////////////////////////////////////
// Mouse Events
//////////////////////////////////////////////////////////////////////////

void AlbumCanvas::OnClick(wxMouseEvent & event)
{
	int x1, y1;
	GetSize(&x1, &y1);

	if (event.LeftDown())
	{
		int toc = (y1-70-14);
		// Start Scrolling, by clicking on the thumb ?
		if (event.GetX()>x1-24 && abs(event.GetY()-toc/2)<=21) 
			StartManualScrolling(event.GetY());

		// Click on one of the nextAlbum/prevAlbum buttons ?
		else if (event.GetX()>x1-24 && event.GetY()>y1-50)
			PageUpDown(event.GetY()<y1-25); // TODO: 25 & 50 are for the 2 scroll buttons

		// Very likely (but not guaranteed) that we clicked on a thumbnail or the album title
		else 
			SelectionClick(event);
	}    
	else if (event.LeftUp() && m_activity==MANUAL_SCROLLING)
	{
		m_pTimer->Stop();
		m_activity = IDLE;
		if (HasCapture())
			ReleaseMouse();
		s_thumb = 0;
		Refresh();
	}
}


void AlbumCanvas::OnRightClick(wxMouseEvent & event)
{
	static wxMenu * pMenu = 0;
//	menuContext { Selection, Album, AllAlbums};

	// Let's build the menu
	m_pController->PopupMenu();
}

void AlbumCanvas::OnDoubleClick(wxMouseEvent& event)
{
	wxBusyCursor wait;

	AlbumView * pAlbum;
	browser::AlbumEntry * pEntry;
	GetEntry(event, pAlbum, pEntry);
	if (!pEntry)
		return;

	m_pAlbumView = pAlbum;
	// m_clipboard.clear(); // TODO: ahem, that's what Picasa does :)
	m_clipboard.insert(pEntry);
	m_pCurrentFile = pEntry;
	currentIndex = pEntry->index;
	m_pController->ShowPicture();
}
class VirtualPaintee : public wxWindow
{
public:
  VirtualPaintee(wxWindow * parent)
    : wxWindow(parent, -1, wxDefaultPosition, wxDefaultSize)
  {
  }
  
};


void AlbumCanvas::MouseMotion(wxMouseEvent& event)
{
	if (m_activity==MANUAL_SCROLLING)
		s_thumb = event.GetY() - s_scroll_start_y;
	else
	{
		AlbumView * pAlbum;
		browser::AlbumEntry * pEntry;
		GetEntry(event, pAlbum, pEntry);
		static browser::AlbumEntry * pPreviousEntry = 0;

		if (pEntry!=pPreviousEntry)
		{
			// On my PC, we need to avoid calling SetCursor too often, or the screen will flicker a bit.
			// Probably a video driver issue, as we're uploading a user-made cursor to the VideoCard
			int cursor = pEntry ? wxCURSOR_HAND : wxCURSOR_ARROW;
			SetCursor(cursor); 
			pPreviousEntry = pEntry;
			
			if (pEntry)
			{
				tm * t = localtime(&pEntry->pFileEntry->pictureDate);
				std::stringstream s;
				s
#ifdef _MSC_VER
				// Watch out! The space here is actually Char(255) .
				  << std::setw(25) << std::left << std::setfill('') 
#endif

				  << pEntry->pFileEntry->name.c_str() << std::right << std::endl 

				  << pEntry->pFileEntry->fileSize << " bytes" << std::endl
				  
				  << std::setw(2) <<  std::setfill('0') << t->tm_mday << "/" 
				  << std::setw(2) <<  std::setfill('0') << t->tm_mon+1 << "/" 
				  << std::setw(2) <<  std::setfill('0') << t->tm_year%100 << " - " 
				  << std::setw(2) <<  std::setfill('0') << t->tm_hour << ":" 
				  << std::setw(2) <<  std::setfill('0') << t->tm_min;

				SetToolTip(s.str().c_str());
			}
			else
				SetToolTip("");
		}
	}
}


void AlbumCanvas::GetEntry(const wxMouseEvent& event, AlbumView * &pAlbumView, browser::AlbumEntry* & pAlbumEntry)
{
	wxClientDC dc(this);
	PrepareDC(dc);	
	wxPoint pos = wxPoint(event.GetX(), m_localScrollPosition + event.GetY());

	pAlbumView =0;
	pAlbumEntry = 0;
	int  currentPos = 0;
	AlbumViews::const_iterator itAlbum = m_albums.begin();
	for (size_t i=0; i<m_albums.size(); ++i)
	{
		AlbumView & album = *m_albums[i];
		if (pos.y >currentPos && pos.y<currentPos+ (signed) album.GetHeight())
		{
			pAlbumView = m_albums[i];
			AlbumView::ClickTarget ct = album.GetClickPosition(pos.x, pos.y-currentPos, pAlbumEntry);
			if (ct!=AlbumView::ctThumbnail)
				pAlbumEntry = 0;
			return;
		}
		currentPos +=album.GetHeight();
	}
}



//////////////////////////////////////////////////////////////////////////
// Scrolling handling
//////////////////////////////////////////////////////////////////////////


void AlbumCanvas::StartManualScrolling(int eventY)
{
	// This is the last chance to get a full refresh, as we'll 
	// disable the Display() method to accelerate things on 
	// the platforms that fancy calling Paint() every time we scroll
	// one pixel
	GetParent()->Refresh(false); 
	Refresh();
	GetParent()->Update();
	Update();
	CaptureMouse();
	m_activity = MANUAL_SCROLLING;
	s_scroll_start_y = eventY;
	wxStartTimer();
	s_timer = wxGetElapsedTime(false);
	m_pTimer->Start(10, wxTIMER_CONTINUOUS);
}


void AlbumCanvas::PageUpDown(bool goingUp)
{
	if (!goingUp)
	{
		for (size_t i=0; i<m_albums.size(); ++i)
			if (signed(m_albums[i]->screen_pos) > m_localScrollPosition+8) // 8 is to get some slack if we're at the top of an album
			{
				JumpTo(m_albums[i]->screen_pos);
				return;
			}
		JumpTo(m_currentHeight - GetSize().y);
	}
	else
	{
		for (int i=m_albums.size()-1; i>=0; --i)
			if (signed(m_albums[i]->screen_pos) < m_localScrollPosition-8) // 8 is to get some slack if we're at the top of an album
			{
				JumpTo(m_albums[i]->screen_pos);
				return;
			}
		JumpTo(0);
	}
}

void AlbumCanvas::JumpTo(int targetY)
{
	targetY = BoundScroll(targetY);
	if (targetY == m_localScrollPosition)
		return;

	s_distance = targetY - m_localScrollPosition;
	s_scroll_start_y = m_localScrollPosition;

	int y1= GetSize().y;
	if (abs(s_distance)>(y1*2+150)) // up to size_y+150 pixels, we scroll
	{ 
		wxBusyCursor bus;
		//so if the distance is too big, we fade to grey..
		m_activity = FADING;
		wxClientDC dc(this);
		int size_y = dc.GetSize().y;
		// On Windows, we're also able to fade nicely
#ifdef USE_FADER
		AlbumFader fader(dc, this, targetY);
		for (int i=0;i<8; ++i)
			fader.Fade(dc, i*2+1);
#endif

		m_localScrollPosition = targetY;

		m_activity = IDLE;
		Refresh();
	}
	else // we scroll
	{
		s_step = 8;
		s_pos = 0;
		m_activity = AUTO_SCROLLING;
		m_pTimer->Start(1, wxTIMER_CONTINUOUS);
		wxStartTimer();
	}

}





int AlbumCanvas::BoundScroll(int y)
{
  int x1, y1;
  GetSize(&x1, &y1);

  // TODO: write a "bound" function
	if (y>(signed)m_currentHeight-y1)
		y= m_currentHeight-y1;

	if (y<0)
		y=0;

	return y;
}

void AlbumCanvas::LocalScroll(int y)
{
  int x1, y1;
  GetSize(&x1, &y1);

  int old_pos = m_localScrollPosition;

  // TODO: write a "bound" function
  m_localScrollPosition= BoundScroll(m_localScrollPosition+y);

  if (old_pos !=m_localScrollPosition)
	{
	  wxRect rect(0,0, x1-24, y1);
	  
#ifdef USE_GTK_PIXBUF
	  _GtkPizza * pPizza = reinterpret_cast <_GtkPizza *> (m_wxwindow);
	  GdkWindow * window = pPizza->bin_window;
	  static GdkGC *gc = gdk_gc_new(window);

	  int dx =0, dy = old_pos-m_localScrollPosition;
	  XCopyArea (GDK_WINDOW_XDISPLAY (window),
		     GDK_WINDOW_XID (window),
		     GDK_WINDOW_XID (window),
		     gdk_x11_gc_get_xgc (gc),
		     rect.x-dx, rect.y - dy,
		     rect.width, rect.height,
		     rect.x, rect.y);
  
#elif WIN32
	  RECT r;
	  r.top = 0; r.left = 0;
	  r.bottom = y1; r.right = x1-24;
	  ::ScrollWindow(GetHwnd(), 0, old_pos-m_localScrollPosition, &r, &r);
#else
	  ScrollWindow(0, old_pos - m_localScrollPosition, &rect);
#endif
	}

}

int  AlbumCanvas::GetLocalScrollPosition() const
{
	return m_localScrollPosition;
}

void AlbumCanvas::OnScrollTimer(wxTimerEvent& event)
{
	if (m_activity==MANUAL_SCROLLING && !HasCapture())
	{
		// 't'is a sad world when you can't even expect a Mouse capture to always function
		m_pTimer->Stop();
		m_activity= IDLE;
		return;
	}
	int x1, y1;
	GetClientSize(&x1, &y1);

	int elapsed = wxGetElapsedTime(false) - s_timer;
	s_timer += elapsed;

	int damage=0;

	if (m_activity == MANUAL_SCROLLING)
	{
		damage = s_thumb*5*elapsed/y1; 
	}
	else if (m_activity == AUTO_SCROLLING)
	{
		double step1 = abs(s_distance)-200.0;
		if (step1<0) step1=0;


 		if (s_pos<step1)
		{
			s_pos+=s_step;
			int shift = (int) (s_pos*abs(s_distance)/s_distance);
			damage = shift+ (s_scroll_start_y- m_localScrollPosition);
		}
 		else if (s_pos<abs(s_distance))
		{
			s_pos+=s_step;
			s_step*=0.965;
			if (s_pos>abs(s_distance))
				s_pos = abs(s_distance);
			int shift = (int) (s_pos*abs(s_distance)/s_distance);
			damage = shift+ (s_scroll_start_y- m_localScrollPosition);
		}
		else
		{
			damage = s_distance + (s_scroll_start_y- m_localScrollPosition);
			m_activity = IDLE;
		}
	}

	LocalScroll (damage);

	wxClientDC dc (this);
#ifdef USE_GTK_PIXBUF
	// We bypass wxWidgets & GTK, so we won't get any expose event.
	// (actually, we could invalidate the region, but that would flicker a bit)
	if (damage<0)
                Display(dc, 0, -damage);
        else if (damage>0)
                Display(dc, y1-damage, y1);
        else // If damage=0, no need to scroll. Just refresh the scrollbar to show the thumb moving
                DrawScrollbar(dc, true);
#else
                DrawScrollbar(dc, true);
#endif
}


//////////////////////////////////////////////////////////////////////////
// Display/Repaint
//////////////////////////////////////////////////////////////////////////


void AlbumCanvas::OnEraseBackGround(wxEraseEvent& event)
{
	//event.Skip();
}
void AlbumCanvas::OnPaint( wxPaintEvent & event )
{
	wxPaintDC dc( this );
	PrepareDC( dc );

	wxRegionIterator upd(GetUpdateRegion()); // get the update rect list

	int min_x = 10000, max_x=0;
  while (upd)
  {
	Display(dc, upd.GetY(), upd.GetY()+upd.GetH());
	if (min_x>upd.GetY())
		min_x = upd.GetY();

	if (max_x<upd.GetY()+upd.GetH())
		max_x = upd.GetY()+upd.GetH();

    upd ++ ;
  }

}


void AlbumCanvas::OnResize(wxSizeEvent & event)
{
  int x1, y1;
  GetSize(&x1, &y1);

  double ratio = ((double)m_localScrollPosition)/m_currentHeight;
	m_currentHeight	= 0;
	for (size_t i=0; i<m_albums.size(); ++i)
	{
		AlbumView * pAlbumView = m_albums[i];
		pAlbumView->OnResize();
		pAlbumView->screen_pos = m_currentHeight;
		m_currentHeight  = m_currentHeight + pAlbumView->GetHeight();
	}

	m_localScrollPosition = int(ratio * m_currentHeight);
	if (m_localScrollPosition>(signed)m_currentHeight-y1)
		m_localScrollPosition = m_currentHeight-y1;

	if (m_localScrollPosition<0)
		m_localScrollPosition=0;

	Refresh();
}

void AlbumCanvas::DrawScrollbar(wxDC & dc, bool fast)
{
	int x, y;
	GetSize(&x, &y);
	static int p_size_y = -1, p_blueSize =-1, p_thumb =-1;

	static bool start = true;
	static wxImage img, imgTop, imgBlue, imgEndBlue, imgEmpty, imgEnd, imgThumb ;
	if (start)
	  {
		 
	         wxMemoryInputStream stream(bmp_phoria3, sizeof bmp_phoria3);
		img = wxImage(stream, wxBITMAP_TYPE_GIF);

		int pos = 0; int height =0;
		pos=0;   height=14; imgTop     = img.GetSubImage(wxRect(0, pos, 24, height));
		pos=14;  height=27; imgBlue    = img.GetSubImage(wxRect(0, pos, 24, 1));
		pos=41;  height=5;  imgEndBlue = img.GetSubImage(wxRect(0, pos, 24, height));
		pos=46;  height=8;  imgEmpty   = img.GetSubImage(wxRect(0, pos, 24, 1));
		pos=54;  height=70; imgEnd     = img.GetSubImage(wxRect(0, pos, 24, height));
		pos=124; height=42; imgThumb   = img.GetSubImage(wxRect(0, pos, 24, height));
		
		start=false;

	  }

	int toc = (y-70-14);
	int blueSize = 0;
	if (m_currentHeight>unsigned(y))
		blueSize = m_localScrollPosition*(toc-2)/(m_currentHeight-y);
	assert (blueSize<toc);

	int thumb= s_thumb+toc/2-21;
	if (thumb<0)
		thumb=0;
	if (thumb+10+42>y-70)
		thumb=y-70-10-42;

	if (fast && p_size_y == y && p_thumb == thumb && p_blueSize == blueSize)
		return;
	p_size_y = y; p_thumb = thumb; p_blueSize = blueSize;
	
	int i =0; const size_t stride = 24*3;

	base::octet * buffer = new base::octet[y*stride];
	memcpy(buffer, imgTop.GetData(), imgTop.GetHeight()*stride);
	base::octet * pTarget = buffer+14*stride;

	while (i<blueSize && i<toc) 
	{
		memcpy(pTarget, imgBlue.GetData(), stride);
		pTarget+=stride;
		++i;
	}
	
	int j=0;
	while (i<toc && j<5)
	{
		memcpy(pTarget, imgEndBlue.GetData()+stride*j++, stride);
		++j; ++i;
		pTarget+=stride;
	}
	
	while (i<toc)
	{
		memcpy(pTarget, imgEmpty.GetData(), stride);
		pTarget+=stride;
		++i;
	}

	memcpy (pTarget, imgEnd.GetData(), imgEnd.GetHeight()*stride);


	pTarget = buffer+(12+thumb)*stride;
	const base::octet * pThumb = imgThumb.GetData();
	base::octet tRed = imgThumb.GetMaskRed(), tGreen = imgThumb.GetMaskGreen(), tBlue = imgThumb.GetMaskBlue();
	for (int i=0;i<imgThumb.GetHeight()*24; ++i)
	{
		if (pThumb[0]==tRed && pThumb[1]==tGreen && pThumb[2] == tBlue)
			pTarget+=3, pThumb+=3;
		else
		{
			*pTarget++ = *pThumb++;
			*pTarget++ = *pThumb++;
			*pTarget++ = *pThumb++;
		}
			
	}

	//dc.DrawBitmap(bmpThumb, x-24, 12+thumb, true);
	wxBitmap bmp (wxImage(24, y, buffer, true));
	dc.DrawBitmap(bmp, x-24, 0);
	delete [] buffer;
}



void AlbumCanvas::Display(wxDC & dc, int min_y, int max_y, int y1)
{
	int y2, dummy;
	GetClientSize(&dummy, &y2);
	const bool fast = !! (m_activity & SCROLLING);
	if (y1==-1)
		y1= GetLocalScrollPosition();

	DrawScrollbar(dc, fast);
	if (max_y==-1)
		max_y = y2;

	unsigned long yStart = min_y+y1, yEnd = max_y+y1;
    unsigned long currentPos=0;

	AlbumViews::const_iterator itAlbum = m_albums.begin();

	while (itAlbum!=m_albums.end() && yStart > currentPos + (*itAlbum)->GetHeight())
	{
		currentPos += (*itAlbum)->GetHeight();
		if (!fast)
			(*itAlbum)->DiscardAlbum();
		itAlbum++;
	}

	while (itAlbum!=m_albums.end() && yEnd > currentPos)
	{
		(*itAlbum)->Display(dc, y1, min_y, max_y);
		currentPos += (*itAlbum)->GetHeight();
		itAlbum++;
	}

	while (itAlbum!=m_albums.end())
	{
		if (!fast)
			(*itAlbum)->DiscardAlbum();
		currentPos += (*itAlbum)->GetHeight();
		itAlbum++;
	}

	if (currentPos<=yEnd)
		AlbumView::FillScreen(dc, currentPos);
}

//////////////////////////////////////////////////////////////////////////
// Albums Handling
//////////////////////////////////////////////////////////////////////////


void AlbumCanvas::LoadAlbum()
{
	wxDirDialog pop (this, "Select the Album", browser::GetDefaultImagePath().c_str(), wxDEFAULT_DIALOG_STYLE /*| wxRESIZE_BORDER*/);

	if (pop.ShowModal()!=wxID_OK )
		return;

	browser::Browser::Instance()->Browse(this, pop.GetPath().c_str());
	m_pGaugeControler->Disable();
}


// Achtung! Non-reentrant code
void AlbumCanvas::BeginBrowsingAlbum(browser::AlbumInfo * pAlbumInfo)
{
  	m_pGaugeControler->Enable("Indexing " + pAlbumInfo->path);
}

void AlbumCanvas::LoadingThumbnail(browser::AlbumInfo * pAlbumInfo, size_t i)
{
	m_pGaugeControler->SetValue((i+1)*100/pAlbumInfo->entries.size());
}


void AlbumCanvas::FinishBrowsingAlbum(browser::AlbumInfo * pAlbumInfo)
{
	if (!pAlbumInfo->entries.empty())
	{
		m_pNextAlbumView = new AlbumView(this, pAlbumInfo);
		m_pNextAlbumView->screen_pos = m_currentHeight;
		m_currentHeight += m_pNextAlbumView->GetHeight();
		m_albums.push_back(m_pNextAlbumView);
	}

	m_pController->AddAlbum(pAlbumInfo);
	Refresh();
}


