#include "stdafx.h"

#include "Browser.h"
#include "Persistence.h"

#include <fstream>
#include <sstream>
#include <string>
#include <sys/stat.h>
#include <wx/dir.h>
#include <wx/timer.h>
#include <wx/filename.h>

#include "JpegDecoder.h"
#include "ExifInfo.h"

using namespace browser;

Browser * Browser::Instance()
{
	static Browser * pInstance = new Browser;
	return pInstance;
}


#ifdef WIN32
#include <shlobj.h>
#include <direct.h>

std::string browser::GetApplicationDirectory()
{
	TCHAR buf[MAX_PATH];
	SHGetFolderPath(0, CSIDL_LOCAL_APPDATA, 0, SHGFP_TYPE_CURRENT, buf);
	return buf + std::string("\\Photophoria");
}

std::string browser::GetDefaultImagePath()
{
	TCHAR buf[MAX_PATH];
	SHGetSpecialFolderPath(NULL, buf, CSIDL_MYPICTURES, NULL);
	return buf;
}

void browser::Mkdir(const char * s)
{
	_mkdir(s);
}
#else

std::string browser::GetApplicationDirectory()
{
	char * home_dir = getenv("HOME");
	if (!home_dir)
		home_dir = ".";

	return home_dir + std::string("/.photophoria");
}

std::string browser::GetDefaultImagePath()
{
	return "/";
}

void browser::Mkdir(const char * s)
{
	int result = mkdir(s, 0750);
	assert(result==0 || errno ==EEXIST);

}

#endif


unsigned long browser::GetUniquePathId(browser::FileEntry * pFileEntry)
{
	unsigned long hash = pFileEntry->pictureDate;
	const char * pName = pFileEntry->path.c_str();
		while (*pName)
		{
			int ch = tolower(*pName++);
			hash += ch + (ch << 17);
			hash ^= hash >> 2;
		}
		return hash;
}




Browser::Browser()
{
    m_exit = false;
	m_pStartMutex = new wxMutex;
	m_pStartCondition = new wxCondition(*m_pStartMutex);
	m_pSemaConsumed = new wxSemaphore();
	m_pSemaProduced = new wxSemaphore();
	wxThreadHelper::Create();
	GetThread()->Run();
	wxSleep(1);

}

Browser::FileType Browser::FileMatch(const std::string & fileName, const struct stat * file_info)
{
	wxString s = wxFileName(fileName.c_str()).GetExt().MakeLower();
	bool canBeJpeg = (s=="jpeg" || s=="jpg");

    // now, is this a normal file ? 
    if (canBeJpeg && (S_IFREG &file_info->st_mode) )
        return  (file_info->st_size<512) ?  Unknown : Picture;

    else if (S_IFDIR & file_info->st_mode)
        return Folder;
    
    return NotRelevant;
}    


void Browser::Browse(BrowserCallback * pCallback, const char * dir, AlbumInfo * pParent)
{
	common::Yield();
	wxDir dirTraverser(dir);

	AlbumInfo * pAlbumInfo = new AlbumInfo(dir, pParent);

	// TODO: Persistence::
	std::vector <std::string> pathsToVisit;

	if (!dirTraverser.IsOpened()) 
	  return;//	throw FileAccessException(std::string("Can't browse ")+dir);

	wxString filename;
    bool cont = dirTraverser.GetFirst(&filename);

	while (cont)
	{
		std::string path (dir);
		path= path + wxFileName::GetPathSeparator() + filename.c_str();
		
		struct stat file_info;
		int error = stat(path.c_str(), &file_info);
		if (!error)
		{
			FileType fileType = FileMatch(path.c_str(), &file_info);
			if (fileType==Folder)
				pathsToVisit.push_back(path);
			else if (fileType==Picture)
			{
				FileEntry * pFileEntry = new FileEntry;
				pFileEntry->pAlbum = pAlbumInfo;
				pFileEntry->path = path;
				pFileEntry->name = filename;
				pFileEntry->fileSize = file_info.st_size;
				pFileEntry->pictureDate = file_info.st_ctime;
				pFileEntry->modificationDate = file_info.st_mtime;
				pFileEntry->guid = GetUniquePathId(pFileEntry);

				// if the file has been allegedly created before 1997
				// then we'll consider its latest modification time instead
				// Note that it's very likely the value will be overriden by the
				// Exif 0x132 tag (Picture DateTime).
				tm * cdate =  localtime(&pFileEntry->pictureDate);
				if (cdate->tm_year < 96)
					pFileEntry->pictureDate = file_info.st_mtime; 
				pAlbumInfo->entries.push_back(pFileEntry);
			}
		}
		cont = dirTraverser.GetNext(&filename);

	}  

	pAlbumInfo->albumId = Persistence::Instance()->GetUniqueId();
//	if (pAlbumInfo->entries.empty() || Persistence::Instance()->AlbumExists(pAlbumInfo)) 
	//	;//delete pAlbumInfo;
	//else
		ThumbnailAlbum(pCallback, pAlbumInfo);

	// And now, let's do some recursivity
	for (size_t i=0; i<pathsToVisit.size(); ++i)
		Browse(pCallback, pathsToVisit[i].c_str(), pAlbumInfo);

}



class JpegDecoderStream : public jpeg_decoder_stream
{
public:
	JpegDecoderStream(uchar * buffer, size_t bufferSize)
		:  m_bufferSize(bufferSize), m_buffer(buffer)
	{
	}

	virtual ~JpegDecoderStream()	{ }
	virtual size_t getSize() { return m_bufferSize;	}
	virtual uchar * getBuffer()	{ return m_buffer; }
	virtual void detach(void) { }
protected:	
	size_t m_bufferSize;
	uchar * m_buffer;
};


static void LoadThumbnail(browser::FileEntry * f)
{
	f->ok=true;
	base::size jpegWidth, jpegHeight;
	base::octet * pRawBuffer = 0;
	try
	{
		JpegDecoderStream input_stream(f->pFileBuffer, f->fileSize);
		jpeg_decoder Pd (&input_stream);

		if (Pd.get_error_code() != 0 || Pd.begin())
			throw common::FileFormatException("Cannot begin JPEG decoding");

		pRawBuffer = Pd.Decode(jpegWidth, jpegHeight);

		if (Pd.get_error_code())
			throw common::FileFormatException("Error during JPEG decoding");;
	}
	catch (...)
	{
		f->ok = false;
	}
	if (pRawBuffer)
		f->thumbnail = new common::Image(jpegWidth, jpegHeight, pRawBuffer, true);
	else
		f->ok = false;
}

static void ReadFileEntry(browser::FileEntry * pEntry)
{
  pEntry->ok = false;
        pEntry->pFileBuffer = 0; // just in case we can't read the file
	std::ifstream f;
	f.open(pEntry->path.c_str(), std::ios::binary);

	if (!f)
	  return;
	
	pEntry->pFileBuffer = new base::octet[pEntry->fileSize + 128];
	
	f.read((char *) pEntry->pFileBuffer, pEntry->fileSize);

	//TODO: this part too should be moved into a padding function
	// we could even allocate the JpegDebugStream here, and give that as a result
	// This would move the file ugliness into the JPEG File encapsulation class
	// where it should always have been...

	unsigned short *ps = reinterpret_cast<unsigned short *> (pEntry->pFileBuffer+pEntry->fileSize);
	size_t n = 64;
	while (n--)
		*ps++ = 0xD9FF;
	pEntry->ok = true; 
}

void Browser::Stop()
{
  m_pStartMutex->Lock();
  m_exit =true;
  m_pStartMutex->Unlock();
  m_pStartCondition->Broadcast();
}

wxThread::ExitCode Browser::Entry()
{
        size_t n = 5; // max number of files in the queue
	m_pStartMutex->Lock();
	while (1)
	{
		m_pStartCondition->Wait();
		if (m_exit)
		  return 0;

		for (size_t i=0; i<m_pCurrentAlbum->entries.size(); ++i)
		{
			FileEntry * pEntry = m_pCurrentAlbum->entries[i];

			ReadFileEntry(pEntry);
			m_pSemaProduced->Post();
			if (i>=n)
				m_pSemaConsumed->Wait();
		}

		if (m_pCurrentAlbum->entries.size()<n)
		  n = m_pCurrentAlbum->entries.size();

		for (size_t j=0; j<n; ++j)
		  m_pSemaConsumed->Wait();
	}
	return 0;
}

void Browser::ThumbnailAlbum(BrowserCallback * pCallback, AlbumInfo * pAlbumInfo) 
{
	pCallback->BeginBrowsingAlbum(pAlbumInfo);

	unsigned int total = pAlbumInfo->entries.size();

	if (total)
	{
		unsigned int decoded = 0;
		Persistence::Transaction transaction;

		assert (total);

		m_pStartMutex->Lock();
		m_pCurrentAlbum = pAlbumInfo;
		m_pStartMutex->Unlock();
		m_pStartCondition->Broadcast();

		while (decoded < total)
		{
			browser::FileEntry * f = pAlbumInfo->entries[decoded];

			m_pSemaProduced->Wait();
			m_pSemaConsumed->Post();
			if (!Persistence::Instance()->HasThumbnail(f))
			{
			  if (f->ok) // If the file was read correctly
				LoadThumbnail(f);
			  if (f->ok)
			    {
			      GetExifTopics(f);
			      
			      Persistence::Instance()->SaveFileEntry(f);
			      Persistence::Instance()->SaveThumbnail(f);
			      delete f->thumbnail;
			      f->thumbnail = 0;
			    }
			}
			delete [] f->pFileBuffer; // Allocated in ReadBuffer, but we ownz it


			pCallback->LoadingThumbnail(pAlbumInfo, decoded);

			++decoded;
		}
		std::vector<FileEntry *>::iterator it = pAlbumInfo->entries.begin();
		while (it!=pAlbumInfo->entries.end())
		{
			if ((*it)->ok)
				++it;
			else
			{
				delete (*it);
				it = pAlbumInfo->entries.erase(it);
			}
		}
		std::sort(pAlbumInfo->entries.begin(), pAlbumInfo->entries.end(), browser::FileEntryTimeLess());


	}		
	Persistence::Instance()->SaveAlbum(pAlbumInfo); 
	pCallback->FinishBrowsingAlbum(pAlbumInfo);
}
