#ifndef Photophoria_Browser_h
#define Photophoria_Browser_h


#include <wx/thread.h>

#include "ExifInfo.h"

struct stat;

namespace browser
{
	struct AlbumInfo;
	struct AlbumEntry;
	struct FileEntry;

	struct FileEntry
	{
		FileEntry ()  // For de-serialisation purposes
			: tagged(false), orientation(0),  thumbnail(0), pFileBuffer(0)
		{
		}

		FileEntry (const std::string & p) 
			: path(p), tagged(false), orientation(0), thumbnail(0), pFileBuffer(0)
		{
		}

		unsigned long guid;

		std::string path;
		std::string name; // constructed from the path
		std::string exposure;
		size_t fileSize;
		time_t modificationDate;
		time_t pictureDate;

		bool tagged;
		int ranking;
		std::string comment;
		int orientation;
		int width, height;

		browser::AlbumInfo * pAlbum;
		common::Image * thumbnail;
		base::octet * pFileBuffer;
		std::vector <ExifTopic> topics;
		bool ok;
	};

	struct FileEntryTimeLess
	{
	bool operator ()(const browser::FileEntry * lhs, const browser::FileEntry * rhs)
	{
		return lhs->pictureDate < rhs->pictureDate;
	}
	};



    struct AlbumEntry
	{
		AlbumEntry() 
			: pFileEntry(0), sysThumbnail(0), selected(false) { }

		FileEntry * pFileEntry;
		bool selected;
		unsigned int index;
		void * sysThumbnail;
		size_t thumb_width;
		size_t thumb_height;
	};

	typedef std::set <browser::AlbumEntry *> ClipboardType;

	struct TagPair
	{
		typedef unsigned short Id;
		TagPair::Id tagNameId;
		TagPair::Id tagValueId;

		std::string tagName;
		std::string tagValue;
	};


	struct AlbumInfo
	{
	public:
		typedef unsigned int Id;
		AlbumInfo() 
			: persisted(false), parentId(0)
		{ 
		}

		AlbumInfo(const char * dir, AlbumInfo * parent)
			: persisted(false), parentId(parent?parent->albumId:0)
		{
			name = dir;
			path = dir;

			std::string::size_type pos = name.find_last_of("/\\");
			if (pos!=std::string::npos)
				name = name.substr(pos+1, std::string::npos);

		}

		~AlbumInfo() 
		{
			for (size_t i=0; i<entries.size(); ++i)
				delete entries[i];
		}


		bool persisted; // transient: indicates that it comes from the DB (otherwise: from the HD)
		AlbumInfo::Id albumId;
		AlbumInfo::Id parentId;

		time_t modificationDate;

		std::string name;
		std::string path;

		std::vector<FileEntry *> entries;
	};


	class FileAccessException : public common::Exception
	{
	public:
		FileAccessException (const std::string &s ) : common::Exception (s){}

	};

	// TODO: The AlbumInfo should be passed const here
	class BrowserCallback
	{
	public:
		virtual browser::FileEntry * GetCurrentFile() = 0;
		virtual void BeginBrowsingAlbum(AlbumInfo *) = 0;
		virtual void LoadingThumbnail(AlbumInfo *, size_t i) = 0;
		virtual void FinishBrowsingAlbum(AlbumInfo *) = 0;
	};


	// General purpose Directory Management
	std::string GetDefaultImagePath();
	std::string GetApplicationDirectory();
	void Mkdir(const char * s);
	unsigned long GetUniquePathId(browser::FileEntry * pFileEntry);



	class Browser : public wxThreadHelper
    {
	enum  FileType { Folder, Zip, Picture, Unknown, NotRelevant};
    public:
   		virtual void Browse(BrowserCallback * pCallback, const char * dir, AlbumInfo * pParentAlbum = 0);
		void Stop(); // Stop the thread
		static Browser * Instance();


    protected:
		Browser::Browser();
		virtual wxThread::ExitCode Entry(); // Entry point for the read-ahead thread

		void ThumbnailAlbum(BrowserCallback * pCallback, AlbumInfo * pAlbumInfo);
		FileType FileMatch(const std::string & fileName, const struct stat * file_info);

		wxMutex *	  m_pStartMutex;
		wxCondition * m_pStartCondition;
		wxSemaphore * m_pSemaConsumed, * m_pSemaProduced;
		AlbumInfo *	  m_pCurrentAlbum;
		bool m_exit; // indicates we want to finish the thread
    };    
}

#endif
