// Based on exiftags, by Eric M. Johnston <emj@postal.net>

#include "stdafx.h"
#include "ExifInfo.h"
#include "Browser.h"

#include "../contrib/exiftags-0.99.1/exif.h"

#include <sstream>
#include <stdio.h>

#define JPEG_M_BEG	0xff	/* Start of marker. */
#define JPEG_M_SOF0	0xc0	/* Start of frame n... */
#define JPEG_M_SOF1	0xc1
#define JPEG_M_SOF2	0xc2
#define JPEG_M_SOF3	0xc3
#define JPEG_M_SOF5	0xc5
#define JPEG_M_SOF6	0xc6
#define JPEG_M_SOF7	0xc7
#define JPEG_M_SOF9	0xc9
#define JPEG_M_SOF10	0xca
#define JPEG_M_SOF11	0xcb
#define JPEG_M_SOF13	0xcd
#define JPEG_M_SOF14	0xce
#define JPEG_M_SOF15	0xcf
#define JPEG_M_SOI	0xd8	/* Start of image. */
#define JPEG_M_EOI	0xd9	/* End of image. */
#define JPEG_M_SOS	0xda	/* Start of scan. */
#define JPEG_M_APP1	0xe1	/* APP1 marker. */
#define JPEG_M_APP2	0xe2	/* APP2 marker. */
#define JPEG_M_ERR	0x100



class ExifException
{
public:
  ExifException (const std::string & s) : what(s) {}
  std::string what;
};

// TODO: it should be a class
static base::offset m_pos;
static base::size m_bufferSize;
static const base::octet * m_buffer;

/*
 * Fetch one byte of the JPEG file.
 */
static int jpg1byte(void)
{
	if (m_pos>=m_bufferSize)
		throw ExifException("invalid JPEG format");
	
	return m_buffer[m_pos++];
}


/*
 * Fetch two bytes of the JPEG file.
 */
static unsigned int jpg2byte(void)
{
	if (m_pos>=m_bufferSize-1)
		throw ExifException("invalid JPEG format");
	
	m_pos+=2;
	return m_buffer[m_pos-2]<<8 | m_buffer[m_pos-1];
}


/*
 * Fetch the length of a marker.
 */
static unsigned int
mkrlen(void)
{
	unsigned int l = jpg2byte();

	/* Length includes itself. */
	if (l< 2)
		throw ExifException("invalid JPEG marker (length mismatch)");
	return (l - 2);
}


/*
 * Skip over an uninteresting marker.
 */
static void
skipmkr(void)
{
	m_pos+=mkrlen();
}


/*
 * Check the first bytes of the file for a marker.
 */
static int topmkr(void)
{
	if (jpg1byte() != JPEG_M_BEG)
		return (JPEG_M_ERR);
	return (jpg1byte());
}


/*
 * Find the next JPEG marker in the file.
 */
static int nxtmkr(void)
{
	int b;
	int bad = 0;

	/* Read until we see JPEG_MARKER. */

	while ((b = jpg1byte()) != JPEG_M_BEG)
		bad++;

	/* Read all JPEG_M_BEGs (which may be used for padding). */

	while ((b = jpg1byte()) == JPEG_M_BEG);

	if (bad)
	  ;// exifwarn("skipped spurious bytes in JPEG");

	return (b);
}


static void sofmrk(browser::FileEntry * pFileEntry)
{
	unsigned long mkrLength = mkrlen();

	int jpg_prcsn = jpg1byte();
	pFileEntry->height = jpg2byte();
	pFileEntry->width = jpg2byte();
}



static const char * GetName(unsigned lvl)
{
	if (lvl== ED_CAM)
			return "Camera Info";
	else if (lvl == ED_IMG)
			return "Image Properties";
	else if (lvl ==ED_BAD )
			return "Bad Properties";
	else if (lvl == ED_OVR)
			return "Overriden Properties";
	else if (lvl == ED_PAS)
			return  "Lens Info";
	else 
		return  "Other Properties";
}


static void GetPictureTime(const char *s, time_t * pt)
{
	struct tm t;
	memset (&t, 0, sizeof t);

	int n = sscanf(s, "%d:%d:%d %d:%d:%d", &t.tm_year, &t.tm_mon, &t.tm_mday, 
										   &t.tm_hour, &t.tm_min, &t.tm_sec);
	t.tm_year -= 1900;
	t.tm_mon -= 1;

	if (n != 6)
		return;
	time_t t2 = mktime(&t);
	if (t2!=-1)
		*pt = t2;
}

static void
AddProps(browser::FileEntry * pFileEntry, struct exifprop *list, unsigned short lvl)
{
	ExifTopic topic(GetName(lvl));

	while (list) 
	{
		if (list->lvl == lvl) 
		{
			const char * n = list->descr ? list->descr : list->name;

			if (list->str)
				topic.properties.push_back(ExifProperty(n, list->str));
			else
			{
				std::stringstream s;
				s << list->value;
				topic.properties.push_back(ExifProperty(n, s.str() ));
			}
			if (list->tag == 0x0100) // Width
				pFileEntry->width = list->value;
			else if (list->tag == 0x0101) // Height
				pFileEntry->height = list->value;
			else if (list->tag == 0x0132 || list->tag == 0x9004 || list->tag == 0x9003) // Time
				GetPictureTime(list->str, &pFileEntry->pictureDate);
			else if (list->tag == 0x0110) // Camera 
				pFileEntry->comment = list->str;
			else if (list->tag == 0x0112) // Orientation
				pFileEntry->orientation = list->value;
			else if (list->tag == 0x829a) // Exposure
				pFileEntry->exposure = list->str;
		}

		list = list->next;
	}
	pFileEntry->topics.push_back(topic);
}


bool GetExifTopics(browser::FileEntry * pFileEntry)
{
	m_pos = 0; 
	m_buffer = pFileEntry->pFileBuffer; 
	m_bufferSize = pFileEntry->fileSize;

	if (topmkr() != JPEG_M_SOI)
		return false; // "start of image not found";

	int count = 0; // number of interesting markers seen so far. We want 2 of them : JPEG_M_APP1 and SOF
	try 
	{

	while(true)
	{
		int mark = nxtmkr();
		switch (mark ) 
		{
			case JPEG_M_APP1:
				{
					unsigned int len = mkrlen();
					struct exiftags *t = exifparse(const_cast<base::octet*> (m_buffer+m_pos), len);

					if (t && t->props) 
					{
						AddProps(pFileEntry, t->props, ED_CAM);
						AddProps(pFileEntry, t->props, ED_IMG);
						AddProps(pFileEntry, t->props, ED_VRB);
					}
					exiffree(t);
					count|=1;
					if (count==3)
						return  false;
					m_pos+=len;
					break;
				}

			case JPEG_M_EOI:
			case JPEG_M_SOS:
				return  false;

			case JPEG_M_SOF0:
			case JPEG_M_SOF1:
			case JPEG_M_SOF3:
			case JPEG_M_SOF5:
			case JPEG_M_SOF6:
			case JPEG_M_SOF7:
			case JPEG_M_SOF9:
			case JPEG_M_SOF10:
			case JPEG_M_SOF11:
			case JPEG_M_SOF13:
			case JPEG_M_SOF14:
			case JPEG_M_SOF15:
				{
					sofmrk(pFileEntry);
					count|=2;
					if (count==3)
						return  false;
					break;	
				}


			case JPEG_M_APP2:
			default:
				skipmkr();
				break;
		}
	}
	}
	catch (ExifException & e)
	{
		common::Debug (e.what);
	}
	return false;
}

