#include <stdafx.h>

#include "Persistence.h"
#include "../contrib/sqlite-3_1_3/sqlite3.h"
#include "../contrib/minilzo/minilzo.h"

#include <map>
#include <algorithm>
#include <iostream>
#include <sstream>
#include <fstream>
#include <errno.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>


std::string TimeSpan(time_t mn, time_t mx);

void GetThumbnailDb(std::string & name, bool & exists)
{
	std::string photo_dir = browser::GetApplicationDirectory();
	name = photo_dir + "/thumbnails.db";
	std::ifstream f;
	f.open(name.c_str(), std::ios::binary);
	exists = !!f;
	if (exists)
	{
		f.close();
	}
	else
	{
		std::cout << "Creating the Album DataBase in " << browser::GetApplicationDirectory() << std::endl;
		browser::Mkdir (photo_dir.c_str());
	}

}

void Persistence::StartDb()
{
  std::string dbName;
  bool exist;
  GetThumbnailDb(dbName, exist);
  int rc = sqlite3_open(dbName.c_str(), &m_pDb); 
  
  
  if( rc )
    std::cout << "Can't open database : " << sqlite3_errmsg(m_pDb) << std::endl;
  else if (!exist)
    {
	RunSql("PRAGMA page_size = 8192");
	RunSql("CREATE TABLE thumbnails (guid INTEGER UNIQUE, width INTEGER, height INTEGER, thumbSize INTEGER, thumbnail BLOB);");
	RunSql("CREATE TABLE fileEntries (guid INTEGER UNIQUE, albumId INTEGER, path TEXT, fileSize INTEGER, modificationDate INTEGER, pictureDate INTEGER, ranking INTEGER, comment TEXT, orientation INTEGER, width INTEGER, height INTEGER);");
	RunSql("CREATE TABLE albums (albumId INTEGER UNIQUE, parentId INTEGER, path TEXT, name TEXT, modificationDate INTEGER);");

	}

  RunSql("PRAGMA synchronous = OFF"); 
}


Persistence * Persistence::Instance()
{
	static Persistence * instance = new Persistence();
	return instance;
}

Persistence::Transaction::Transaction()
{
	Persistence::Instance()->RunSql("BEGIN;");
}
Persistence::Transaction::~Transaction()
{
	Persistence::Instance()->RunSql("COMMIT;");
}



Persistence::Persistence()
{
	m_currentId = 1;
	StartDb();
}


Persistence::~Persistence ()
{
	//close the DB
	sqlite3_close(m_pDb);

	// and get rid of all objects
	for (size_t i=0; i<m_albums.size(); ++i)
		delete m_albums[i];

} 


static int StaticThumbnailCallback(void *pEntry, int argc, char **argv, char **azColName)
{
	return 0;
}

sqlite3_stmt * Persistence::PrepareStatement (const std::string & s)
{
	const char * pDummy;
	sqlite3_stmt *pStmt = 0;
	int rc = sqlite3_prepare(m_pDb, s.c_str(), s.size(), &pStmt, &pDummy);
	if (rc!=SQLITE_OK)
	  {
	    std::cout << "Error while Preparing SQL "<< std::endl;
	    std::cout << "   QUERY="<<s<<std::endl;
	    std::cout << "   ERROR="<< sqlite3_errmsg(m_pDb) << std::endl;
	  }
	assert(rc==SQLITE_OK);
	return pStmt;
}


void Persistence::RunSql(const std::string & s, bool ignoreErrors)
{
	char *zErrMsg = 0;
	int rc = sqlite3_exec(m_pDb, s.c_str(), StaticThumbnailCallback, 0, &zErrMsg);
	if( rc!=SQLITE_OK && !ignoreErrors)
	{
		common::Debug(std::string("SQL error: ") +  zErrMsg);
		BREAKPOINT;
		//TODO:THROW
	}
}

void Persistence::LoadThumbnail(browser::FileEntry * pFileEntry )
{
	static sqlite3_stmt *pStmt = PrepareStatement("SELECT width, height, thumbSize, thumbnail FROM thumbnails WHERE guid=?;");

	int rc = sqlite3_bind_int(pStmt, 1, (signed) pFileEntry->guid);

	rc = sqlite3_step(pStmt);
	assert(rc==SQLITE_DONE || rc==SQLITE_ROW);
	
	if (rc==SQLITE_ROW)
	{
		int width = sqlite3_column_int(pStmt, 0);
		int height = sqlite3_column_int(pStmt, 1);
		int len = sqlite3_column_int(pStmt, 2);
		const unsigned char * pCompressedBuffer = 
			static_cast <const unsigned char *> (sqlite3_column_blob(pStmt, 3));

		// decompression with miniLzo
		unsigned int bufferSize = width*height*3;
		base::octet * buffer = new base::octet[bufferSize];
		lzo1x_decompress(pCompressedBuffer, len, buffer, &bufferSize, 0);
		assert(bufferSize == size_t(width*height*3));

		// get it.
		pFileEntry->thumbnail = new common::Image(width, height, buffer, true);
	}

	sqlite3_reset(pStmt);
}

bool Persistence::HasThumbnail(browser::FileEntry * pFileEntry )
{
	static sqlite3_stmt *pStmt = PrepareStatement("SELECT guid FROM thumbnails WHERE guid=?");

	int rc = sqlite3_bind_int(pStmt, 1, (signed) pFileEntry->guid);

	rc = sqlite3_step(pStmt);
	assert(rc==SQLITE_DONE || rc==SQLITE_ROW);
	sqlite3_reset(pStmt);
	return (rc==SQLITE_ROW);
}


void Persistence::SaveThumbnail(browser::FileEntry * pFileEntry)
{
	if (!pFileEntry->thumbnail)
		return;

	static sqlite3_stmt *pStmt = 
		PrepareStatement ("INSERT INTO thumbnails (guid, width, height, thumbSize, thumbnail) VALUES(?, ?, ?, ?, ?);");

	// compress with miniLZO
	base::size bufferSize = pFileEntry->thumbnail->width*pFileEntry->thumbnail->height*3;
	unsigned int len = bufferSize*3/2;
	static unsigned char * workerMem = new base::octet[LZO1X_1_MEM_COMPRESS];
	base::octet * compressedBuffer= new base::octet[len]; // 50% ought to be enough
	lzo1x_1_compress (pFileEntry->thumbnail->buffer, bufferSize, 
							   compressedBuffer, &len, workerMem);

	
	sqlite3_bind_int(pStmt, 1, pFileEntry->guid);
	sqlite3_bind_int(pStmt, 2, pFileEntry->thumbnail->width);
	sqlite3_bind_int(pStmt, 3, pFileEntry->thumbnail->height);
	sqlite3_bind_int(pStmt, 4, len);
	sqlite3_bind_blob(pStmt, 5, compressedBuffer, len, SQLITE_TRANSIENT);

	int rc = sqlite3_step(pStmt);
	if( rc!=SQLITE_DONE)
	{
	  std::cerr << "SQL error on " << pFileEntry->path << " : " << rc << std::endl;
		BREAKPOINT;
		//TODO:THROW
	}
	sqlite3_reset(pStmt);
}



void Persistence::LoadAlbums()
{
	if (!m_albums.empty())
		return;

	static sqlite3_stmt *pStmt = PrepareStatement("SELECT albumId, parentId, path, name, modificationDate FROM albums;");

	int rc = sqlite3_step(pStmt);
	assert(rc==SQLITE_DONE || rc==SQLITE_ROW);
	
	while (rc==SQLITE_ROW)
	{
		browser::AlbumInfo * pAlbum = new browser::AlbumInfo;
		m_albums.push_back(pAlbum);
		pAlbum->persisted = true;
		pAlbum->albumId = sqlite3_column_int(pStmt, 0);
		pAlbum->parentId = sqlite3_column_int(pStmt, 1);
		pAlbum->path = (const char * ) sqlite3_column_text(pStmt, 2);
		pAlbum->name = (const char * ) sqlite3_column_text(pStmt, 3);
		pAlbum->modificationDate = sqlite3_column_int(pStmt, 4);

		if (pAlbum->albumId >= m_currentId)
			m_currentId = pAlbum->albumId+1;

		rc = sqlite3_step(pStmt);
	}
	assert(rc==SQLITE_DONE);

	sqlite3_reset(pStmt);

	for (size_t i=0; i<m_albums.size(); ++i)
		LoadFileEntries(m_albums[i]);
}


void Persistence::LoadFileEntries(browser::AlbumInfo * pAlbumInfo)
{
	static sqlite3_stmt *pStmt = PrepareStatement("SELECT guid, path, fileSize, modificationDate, pictureDate, ranking, comment, orientation, width, height FROM fileEntries WHERE albumId=?;");

	sqlite3_bind_int(pStmt, 1, (signed) pAlbumInfo->albumId);

	int rc = sqlite3_step(pStmt);
	assert(rc==SQLITE_DONE || rc==SQLITE_ROW);
	
	while (rc==SQLITE_ROW)
	{
		browser::FileEntry * pEntry = new browser::FileEntry;

		pEntry->pAlbum = pAlbumInfo;
		pEntry->guid = sqlite3_column_int(pStmt, 0);
		pEntry->path = (const char *) sqlite3_column_text(pStmt, 1);
		pEntry->fileSize = sqlite3_column_int(pStmt, 2);
		pEntry->modificationDate = sqlite3_column_int(pStmt, 3);
		pEntry->pictureDate = sqlite3_column_int(pStmt, 4);
		pEntry->ranking = sqlite3_column_int(pStmt, 5);
		pEntry->comment =  (const char *) sqlite3_column_text(pStmt, 6);
		pEntry->orientation = sqlite3_column_int(pStmt, 7);
		pEntry->width = sqlite3_column_int(pStmt, 8);
		pEntry->height = sqlite3_column_int(pStmt, 9);

		std::string::size_type pos = pEntry->path.find_last_of("/\\");
		if (pos!=std::string::npos)
				pEntry->name = pEntry->path.substr(pos+1, std::string::npos);

		pAlbumInfo->entries.push_back(pEntry);   

		 rc = sqlite3_step(pStmt);
	}

	assert(rc==SQLITE_DONE);

	sqlite3_reset(pStmt);
}

void Persistence::SaveFileEntry(browser::FileEntry * pFileEntry)
{
	assert (pFileEntry);

	static sqlite3_stmt *pStmt = 
		PrepareStatement ("INSERT INTO fileEntries (guid, albumId, path, fileSize, modificationDate, pictureDate, ranking, comment, orientation, width, height) VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?);");

	sqlite3_bind_int (pStmt, 1, pFileEntry->guid);
	sqlite3_bind_int (pStmt, 2, pFileEntry->pAlbum->albumId);
	sqlite3_bind_text(pStmt, 3, pFileEntry->path.c_str(), pFileEntry->path.size(), SQLITE_TRANSIENT);
	sqlite3_bind_int (pStmt, 4, pFileEntry->fileSize);
	sqlite3_bind_int (pStmt, 5, pFileEntry->modificationDate);
	sqlite3_bind_int (pStmt, 6, pFileEntry->pictureDate);
	sqlite3_bind_int (pStmt, 7, pFileEntry->ranking);
	sqlite3_bind_text(pStmt, 8, pFileEntry->comment.c_str(), pFileEntry->comment.size(), SQLITE_TRANSIENT);
	sqlite3_bind_int (pStmt, 9, pFileEntry->orientation);
	sqlite3_bind_int (pStmt,10, pFileEntry->width);
	sqlite3_bind_int (pStmt,11, pFileEntry->height);

	int rc = sqlite3_step(pStmt);
	if( rc!=SQLITE_DONE)
	{
	  std::stringstream s;
	  s << "SQL error on " << pFileEntry->path << " : " << rc << std::endl;
	  common::Debug(s.str());
		BREAKPOINT;
		//TODO:THROW
	}
	sqlite3_reset(pStmt);
}


void Persistence::SaveAlbum(browser::AlbumInfo * pAlbum)
{
	assert (pAlbum);
	if (pAlbum->persisted)
		return;
	pAlbum->persisted = true;
	m_albums.push_back(pAlbum);

	static sqlite3_stmt *pStmt = 
		PrepareStatement ("INSERT INTO albums (albumId, parentId, path, name, modificationDate) VALUES(?, ?, ?, ?, ?);");

	sqlite3_bind_int (pStmt, 1, pAlbum->albumId);
	sqlite3_bind_int (pStmt, 2, pAlbum->parentId);
	sqlite3_bind_text(pStmt, 3, pAlbum->path.c_str(), pAlbum->path.size(), SQLITE_TRANSIENT);
	sqlite3_bind_text(pStmt, 4, pAlbum->name.c_str(), pAlbum->name.size(), SQLITE_TRANSIENT);
	sqlite3_bind_int (pStmt, 5, pAlbum->modificationDate);

	int rc = sqlite3_step(pStmt);
	if( rc!=SQLITE_DONE)
	{
	  std::cerr << "SQL error on " << pAlbum->path << " : " << rc << std::endl;
		BREAKPOINT;
		//TODO:THROW
	}
	sqlite3_reset(pStmt);
}



//  FileEntryMagicLess(&FileEntry::pictureDate);
template <typename T>
struct FileEntryMagicLess
{
	FileEntryMagicLess(T browser::FileEntry::*member) : m(member) { };
	T browser::FileEntry::*m;


	bool operator ()(const browser::FileEntry * lhs, const browser::FileEntry * rhs)
	{
		return lhs->*m < rhs->*m;
	}
};

const Persistence::DateAlbums & Persistence::GetMonths()
{

	static DateAlbums monthlyAlbums;
	if (monthlyAlbums.empty())
	{
		for (size_t i=0; i<m_albums.size(); ++i)
			for (size_t j =0; j<m_albums[i]->entries.size(); ++j)	
			{
				browser::FileEntry * pFileEntry = m_albums[i]->entries[j];

				tm * minDate =  localtime(&pFileEntry->pictureDate);
				int month = minDate->tm_year*12 + minDate->tm_mon;

				browser::AlbumInfo * & pAlbum = monthlyAlbums[month];
				if (!pAlbum)
				{
					const char * monthName [] = { "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December" };
					pAlbum = new browser::AlbumInfo();
					std::stringstream s;
					s << monthName[month%12] << " " << month/12 + 1900;
					pAlbum->name = s.str();
				}
				pAlbum->entries.push_back(pFileEntry);
			}

		DateAlbums::const_iterator it = monthlyAlbums.begin(),
							    itEnd = monthlyAlbums.end();

		while (it != itEnd)
		{
			browser::AlbumInfo * pAlbum = it->second;
			std::sort(pAlbum->entries.begin(), pAlbum->entries.end(), FileEntryMagicLess<time_t>(&browser::FileEntry::pictureDate));
			++it;
		}
	}


	return monthlyAlbums;
}


const std::vector <browser::AlbumInfo * > &  Persistence::GetCalendarAlbums()
{
	static std::vector <browser::AlbumInfo * > albums;
	if (albums.empty())
	{
		const DateAlbums & monthlyAlbums = GetMonths();
		DateAlbums::const_iterator it = monthlyAlbums.begin(),
							    itEnd = monthlyAlbums.end();
		
		while (it != itEnd)
		{
			browser::AlbumInfo * pAlbum = it->second;
			albums.push_back(pAlbum);
			++it;
		}
	}
	return albums;
}

const std::vector<browser::AlbumInfo *> & Persistence::GetAllAlbums() 
{
	if (m_albums.empty())
		LoadAlbums();
	return m_albums;
}

const std::vector<browser::AlbumInfo *> & Persistence::GetAlbums(Persistence::Ordering order)
{
	if (order==CALENDAR)
		return GetCalendarAlbums();
	else
		return GetAllAlbums();
}


bool Persistence::AlbumExists(browser::AlbumInfo * pAlbum)
{
	for (size_t i=0; i <m_albums.size(); ++i)
		if (m_albums[i]->path == pAlbum->path)
			return true;
	return false;
}

browser::AlbumInfo::Id Persistence::GetUniqueId()
{
	return m_currentId++;
}

