#include "stdafx.h"
#include "common.h"
#include "gaussian.h"
#include <cmath>

using namespace transform;



void Exposition::DoHighlightsShadows (const common::Image & image, common::Image & image2, 
							  base::param shadow_cutoff, base::param highlight_cutoff, 
							  base::param gamma1, base::param gamma2)
{
	base::size buf_size = image.width*(image.height-1)-1;

	base::octet * p1 = image.buffer;
	base::octet * p2 = image2.buffer;
	for (base::offset i=0; i<buf_size; ++i, p1+=3, p2+=3)
	{

		base::octet r=p1[base::cRed], g=p1[base::cGreen], b=p1[base::cBlue];
		double blurry  = GetY(p2[base::cRed], p2[base::cGreen], p2[base::cBlue])/255.0;
		double focused = GetY(r, g, b)/255.0;

		if (focused > highlight_cutoff)
		{
			// convert focus on a [0,1]
			double focus_normalised = (focused-highlight_cutoff)/(1-highlight_cutoff);

			// apply the local gamma correction
			double new_y = pow(focus_normalised, (double)gamma2)*(1-highlight_cutoff)+highlight_cutoff;

			// and apply the unsharp mask
			if (focused>blurry)
				new_y+=focus_normalised*0.5*(focused-blurry);

			Recolor(focused*255, new_y*255, r, g, b);
		}

		if (focused<shadow_cutoff)
		{
			double focus_normalised = focused/shadow_cutoff;
			// apply the local gamma correction
			double new_y = pow(focus_normalised, 1.0/gamma1)*shadow_cutoff;

			// and apply the unsharp mask
			new_y+= (1-focus_normalised)*0.5*(focused-blurry);
			Recolor(focused*255, new_y*255, r, g, b);
		}

		p2[base::cRed]=r; p2[base::cGreen]=g; p2[base::cBlue]=b;
	}

}

void UnsharpMask::DoUnsharpMask (const common::Image & image, common::Image & image2, const base::param c)
{
	for (base::offset i=0; i<image.width*image.height*3; ++i)
		image2.buffer[i] = clamp(image.buffer[i]*(1+c) - image2.buffer[i]*c);
}




void GaussianBlur::DoGaussian(GaugeControler * pGaugeControler, const common::Image & image, common::Image & image2, base::param n)
{
	base::octet * buffer = new base::octet[3*image.width*image.height];
	common::Image img_temp(image.width, image.height, buffer, true);

	CreateConvolutionVector(n);
	ConvoluteDimension(pGaugeControler, image, img_temp, false);
	ConvoluteDimension(pGaugeControler, img_temp, image2, true);
}

void GaussianBlur::CreateConvolutionVector (base::param radius)
{

	// d is the effective diameter in pixels; all weight factors outside d are assumed
	// to be zero. The factor 5.0 is somewhat arbitrary; a value between 4.0 and 6.0
	// is generally recommended.
	base::size d = static_cast <int> (5.0f * radius + 1.0f);

	d |= 1;		// must be odd
    m_dim = d;
	if (d <= 1) return ;	// radius 0 is acceptable; effectively no convolution
 
    m_convolutionVector = new gaussianParam[d];
	d /= 2;

	base::param num = 2 * radius * radius;
	base::param f = expf(d * d / num);
	m_denominator = static_cast <gaussianParam> (f);

	m_convolutionVector[d] = m_denominator;

	for (base::offset i = 1; i <= d; i++)
	{
		int i2 = - (int)(i * i);
		gaussianParam v = static_cast <gaussianParam> (f * expf(i2 / num));
		m_convolutionVector[d - i] = v;
		m_convolutionVector[d + i] = v;
		m_denominator += 2 * v;
	}


}



// horizontal direction, or in vertical direction (depending on bHorizontal).
void GaussianBlur::ConvoluteDimension(GaugeControler * pGaugeControler, const common::Image & image, common::Image & image2, bool bHorizontal)
{
	if (m_dim <= 1) 
		return;
	unsigned int d = m_dim / 2;
	

	unsigned int nLines;		// number of lines (horizontal or vertical)
	unsigned int nPixels;		// number of pixels per line
	unsigned int dPixelSrc;		// pixel step in source
	unsigned int dLineSrc;		// line step in source

	if (bHorizontal)
	{
		nLines = image.height;
		nPixels = image.width;
		dPixelSrc = 3;
		dLineSrc = image.width*3;
	}
	else
	{
		nLines = image.width;
		nPixels = image.height;
		dPixelSrc = image.width*3;
		dLineSrc = 3;
	}

	// This line added in version 1.1: avoid overrun in small bitmaps.
	if (d > nPixels / 2) d = nPixels / 2;

	for (int plane = 0; plane < 3; plane++)	// loop through color planes
	{
		if (pGaugeControler)
			pGaugeControler->SetValue(8+plane*17+(bHorizontal?50:0));
		base::octet * pLineSrc = image.buffer+plane;
		base::octet * pLineDest = image2.buffer+plane;
		
		for (base::offset line = 0; line < nLines; line++)	// loop through lines
		{
			base::octet * pPixelDest = pLineDest;

			base::offset pxl = 0;
			for (pxl = 0; pxl < d; pxl++)	// loop through pixels in left/top margin
			{
				gaussianParam * pFactors = m_convolutionVector + d - pxl;

				base::size xEnd = pxl + d;
				if (xEnd > nPixels) xEnd = nPixels;

				gaussianParam sum = 0, denom=0;

				base::octet * pPixelSrc = pLineSrc;

				for (base::offset x = 0; x < xEnd; x++)
				{
					denom += *pFactors;
					sum += *pFactors++ * *pPixelSrc;
					pPixelSrc += dPixelSrc;
				}

				if (denom) sum /= denom;
				* pPixelDest = (base::octet) sum;

				pPixelDest += dPixelSrc;
			}

			for (pxl = d; pxl < nPixels - d; pxl++)	// loop through pixels in main area
			{
				gaussianParam * pFactors = m_convolutionVector;
				gaussianParam sum = 0;

				base::offset xBegin = pxl - d;
				base::octet * pPixelSrc = & pLineSrc[xBegin * dPixelSrc];

				for (base::offset x = xBegin; x <= pxl + d; x++)
				{
					sum += *pFactors++ * *pPixelSrc;
					pPixelSrc += dPixelSrc;
				}
				sum /= m_denominator;

				* pPixelDest = static_cast<base::octet> (sum);

				pPixelDest += dPixelSrc;
			}

			for (pxl = nPixels - d; pxl < nPixels; pxl++)
				// loop through pixels in right/bottom margin
			{
				gaussianParam * pFactors = m_convolutionVector;
				gaussianParam denom = 0;
				gaussianParam sum = 0;

				int xBegin = pxl - d;
				if (xBegin < 0) xBegin = 0;
				base::octet * pPixelSrc = & pLineSrc[xBegin * dPixelSrc];

				for (base::offset x = xBegin; x < nPixels; x++)
				{
					denom += *pFactors;
					sum += *pFactors++ * *pPixelSrc;
					pPixelSrc += dPixelSrc;
				}

				if (denom) sum /= denom;

				* pPixelDest = (base::octet) sum;

				pPixelDest += dPixelSrc;
			}
			pLineSrc += dLineSrc;
			pLineDest += dLineSrc;
		}	// next line
	

	}	// next plane
}

