#ifndef Photophoria_transform_h
#define Photophoria_transform_h

#include "common.h"
#include "math.h"
namespace transform
{
    //////////////////////////////////////////////:
    // Control, Parameter and Transform, used to exchange infos about the 
    // transform with the GUI
    //////////////////////////////////////////////:
    struct Control
    {
    	enum ControlType { FLOAT_VALUE, INTEGER_SLIDER, FLOAT_SLIDER };

        Control (const std::string & n, ControlType , base::param def)
         : name(n),  defaultValue(def) {}
         
        std::string name;
		base::param defaultValue;
    };
    typedef std::vector<Control> Controls;
   
     
    struct Parameter
    {
		Parameter (base::param p) : value(p) {}
        base::param value;
    };    
    typedef std::vector<Parameter> Parameters;
    
    

	inline int clamp (int i)
	{
		if (i & 0xFFFFFF00)
			i = (((~i) >> 31) & 0xFF);

		return (i);
	}

	inline int clamp (double d)
	{
		return clamp(static_cast<int>(d));
	}




    class Transform
    {
    public:
        virtual const Controls & GetControls() const = 0;
		virtual bool Execute(GaugeControler * pGaugeControler, 
							 const common::Image & in, common::Image & out, 
  						     const transform::Parameters & params) = 0; 
        virtual ~Transform() { } // we don't make it virtual pure, so that we 
        						 // don't force the derived classes to have one.
		unsigned long userData;

		// TODO: this goes into Color.h of course
		static inline void Recolor2(int y, int new_y, double saturation, base::octet & r, base::octet & g, base::octet &b, bool no=false)
		{
			//	int y  = clamp( 0.2990*r + 0.5870*g + 0.1140*b);
			int cb = clamp(-0.1687*r - 0.3313*g + 0.5000*b +128);
			int cr = clamp(  0.500*r - 0.4187*g - 0.0813*b +128);

			//	change the saturation
			if (no && y<new_y)
				saturation *= (double)y/new_y;
			else if (no &&  y>new_y)
				saturation *= (double)new_y/y; 

			r = clamp(new_y + saturation*(1.402 * (cr-128.0)) );
			g = clamp(new_y + saturation*(-0.34414 *(cb-128.0) - 0.71414*(cr-128.0)) );
			b = clamp(new_y + saturation*(1.772 * ( cb -128.0)));
		}
		static inline void Recolor2(double y, double new_y, double saturation, base::octet & r, base::octet & g, base::octet &b, bool no=false)
		{
			Recolor2(static_cast<int>(y), static_cast<int>(new_y), saturation, r, g, b, no);
		}

		static inline void Recolor(int y, int new_y, base::octet & r, base::octet & g, base::octet &b)
		{
			Recolor2(y, new_y, 1.0, r, g, b);
		}
		static inline void Recolor(double y, double new_y, base::octet & r, base::octet & g, base::octet &b)
		{
			Recolor2(static_cast<int>(y), static_cast<int>(new_y), 1.0, r, g, b);
		}
		static inline int GetY(base::octet r, base::octet g, base::octet b)
		{
			return clamp( 0.2990*r + 0.5870*g + 0.1140*b);
		}

		static void RGB2YCbCr(common::Image & image)
		{

			base::octet * p = image.buffer;
			for (base::offset x=image.height*image.width; x; --x)
			{
				base::octet r=p[base::cRed], g=p[base::cGreen], b=p[base::cBlue];

				base::octet y  = clamp( 0.2990*r + 0.5870*g + 0.1140*b);
				base::octet cb = clamp(-0.1687*r - 0.3313*g + 0.5000*b +128);
				base::octet cr = clamp(  0.500*r - 0.4187*g - 0.0813*b +128);

				// regardless of the display settings, we always output Y Cb Cr in this order
				p[0]=y; p[1]=cb; p[2]=cr;
				p+=3;

			}

		}

		static void YCbCr2RGB(common::Image & image)
		{

			base::octet * p = image.buffer;
			for (base::offset x=image.height*image.width; x; --x)
			{
				base::octet y=p[0], cb=p[1], cr=p[2];

				base::octet r = clamp(y + 1.402 * (cr-128.0) );
				base::octet g = clamp(y -0.34414 *(cb-128.0) - 0.71414*(cr-128.0) );
				base::octet b = clamp(y + 1.772 * ( cb -128.0));


				p[base::cRed]=r; p[base::cGreen]=g; p[base::cBlue]=b;
				p+=3;
			}
		}


    };
    
    


    //////////////////////////////////////////////:
    // Brightness / Contrast
    //////////////////////////////////////////////:
    class BrightnessContrast : public Transform
    {
    public:
        virtual const Controls & GetControls() const
        {
            static Controls controls;
            if (controls.empty())
            {
                controls.push_back(Control("Brightness", Control::FLOAT_SLIDER, 0.5));
				controls.push_back(Control("Contrast",	 Control::FLOAT_SLIDER, 0.5));
				controls.push_back(Control("Saturation", Control::FLOAT_SLIDER, 0.5));
            }
            return controls;    
        };
        
		virtual bool Execute(GaugeControler * pGaugeControler, 
							 const common::Image & in, common::Image & out, 
			       			 const transform::Parameters & params)
        {
			base::param brightness = params[0].value;
			base::param contrast = params[1].value;
			base::param saturation = params[2].value;
          
            DoBC(in, out, brightness, contrast, saturation*2);
            
            return true; // cannot fail
        }    
        
        void DoBC (const common::Image & in, common::Image & out, 
					base::param brightness, base::param contrast, base::param saturation) const;

    protected:
    };    


   template <typename LightTrait>
   class ChangeBrightness : public Transform
    {
    public:
        virtual const Controls & GetControls() const
        {
            static Controls controls;
            if (controls.empty())
            {
                controls.push_back(Control("Brightness", Control::FLOAT_SLIDER, 0.5));
            }
            return controls;    
        };
        
		virtual bool Execute(GaugeControler * pGaugeControler, 
							 const common::Image & in, common::Image & out, 
			       			 const transform::Parameters & params)
        {
			base::param brightness = params[0].value*4;
     
			base::octet * p1 = in.buffer;
			base::octet * p2 = out.buffer;

			for (base::offset i=0; i<in.width*in.height; ++i, p1+=3, p2+=3)
			{
				base::octet r = p1[base::cRed], g=p1[base::cGreen], b=p1[base::cBlue]; 
				base::param y = GetY(r,g,b);
				base::param newy = y*brightness;
				LightTrait::Recolor(y, newy, r,g,b);

				p2[base::cRed] = r;  p2[base::cGreen] = g;  p2[base::cBlue] = b;
			}

			return true; // cannot fail
		}
		

    protected:
    };    


    //////////////////////////////////////////////:
    // Auto-Correct
    //////////////////////////////////////////////:
    class AutoCorrect : public Transform
    {
    public:
        virtual const Controls & GetControls() const
        {
            static Controls controls;
            if (controls.empty())
            {
                controls.push_back(Control("Brightness", Control::FLOAT_SLIDER, 1.0f));
                controls.push_back(Control("Color",	     Control::FLOAT_SLIDER, 0.8f));
            }
            return controls;    
        };
        
		virtual bool Execute(GaugeControler * pGaugeControler, 
							 const common::Image & in, common::Image & out, 
			       			 const transform::Parameters & params)
        {
			base::param b = params[0].value;
            base::param c = params[1].value;
          
            Correct(in, out, b, c);
            
            return true; // cannot fail
        }    
        
        void Correct(const common::Image & in, common::Image & out, base::param b, base::param c) const;

    protected:
    };    










    //////////////////////////////////////////////:
    // Resize using Bilinear Filtering
    //////////////////////////////////////////////:

    class BilinearResize : public Transform
    {
    public:
        virtual const Controls & GetControls(const common::Image &) const
        {
            static Controls controls;
            if (controls.empty())
            {
                controls.push_back(Control("width",  Control::INTEGER_SLIDER, 8.0));
                controls.push_back(Control("height", Control::INTEGER_SLIDER, 8.0));
            }
            return controls;    
        };
        
		virtual bool Execute(GaugeControler * pGaugeControler, 
						     const common::Image & in, common::Image & out, 
	       					 const transform::Parameters & params)
        {
            Resize(in, out);
            return true; // cannot fail
        }    
                    
        static void Resize(const common::Image & in, common::Image & out);    

    protected:
        static inline base::octet Bilinear(const base::octet * p, base::size width, 
         								   base::octet orig_offset_x, base::octet orig_offset_y );
    };    

  
} 
   
#endif

